"""
    String conditions base class
"""

import datetime
import logging

from dateutil.parser import parse
from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


class MonthDayCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    class DateFormat:
        DAY_FORMAT = "%m-%d"
        ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    def __init__(self, values, what, date_format=DateFormat.ISO_FORMAT, qualifier=ConditionBase.Qualifier.ForAnyValue,
                 **kwargs):
        self.values = values
        self.what = what
        self.qualifier = qualifier
        self.date_format = date_format

    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()

    def convert_str_to_date(self, d_str):
        try:
            return datetime.datetime.strptime(d_str, self.DateFormat.ISO_FORMAT)
        except:
            return None

    def check_format_input_day(self, i):
        if isinstance(i, str):
            if self.convert_str_to_date(i):
                return True
        return False

    def convert_string_to_day_format(self, date_str):
        return datetime.datetime.strptime(date_str, self.date_format).strftime(self.DateFormat.DAY_FORMAT)

    def convert_date_to_day_format(self, d):
        return d.strftime(self.DateFormat.DAY_FORMAT)

    def convert_timestamp_to_day_format(self, timestamp):
        # try:
        return datetime.datetime.fromtimestamp(timestamp).strftime(self.DateFormat.DAY_FORMAT)
        # except:
        #     return None

    def convert_string_to_day_format_from_parse(self, date_str):
        try:
            return parse(date_str).strftime(self.DateFormat.DAY_FORMAT)
        except:
            return None

    def convert_day_format_from_any(self, i):
        if isinstance(i, (int, float)):
            day_format = self.convert_timestamp_to_day_format(i)
        elif isinstance(i, str):
            day_format = self.convert_string_to_day_format_from_parse(i)
            if day_format is None:
                day_format = self.convert_string_to_day_format(i)
        elif isinstance(i, datetime.datetime):
            day_format = self.convert_date_to_day_format(i)
        else:
            return False
        return day_format


class MonthDayConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.Raw(required=True, allow_none=True), required=True, allow_none=False)
    what = fields.Raw(required=True, allow_none=True)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                              validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    date_format = fields.String(allow_none=False, load_default=MonthDayCondition.DateFormat.ISO_FORMAT, )

    class Meta:
        unknown = EXCLUDE



"""
    String not contains conditions
"""

from marshmallow import post_load

from .base import StringCondition, StringConditionSchema


class NotContains(StringCondition):
    """
        Condition for string `self.what` not contains `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if not self.is_string(i) and not self.is_string(self.what):
                    continue
                elif not self.is_string(i) or not self.is_string(self.what):
                    return True

                if self.delimiter:
                    if i + self.delimiter in self.what or i == self.what:
                        return False
                else:
                    if i in self.what:
                        return False
            return True
        else:
            for i in self.values:
                if not self.is_string(i) and not self.is_string(self.what):
                    return False
                elif not self.is_string(i) or not self.is_string(self.what):
                    continue

                if self.delimiter:
                    if i + self.delimiter in self.what or i == self.what:
                        return False
                else:
                    if i in self.what:
                        return False
            return True


class NotContainsSchema(StringConditionSchema):
    """
        JSON schema for not contains string condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return NotContains(**data)



"""
    Numeric not equal conditions
"""

from marshmallow import post_load

from .base import DateCondition, DateConditionSchema


class DateNeq(DateCondition):
    """
        Condition for number `what` not equals `value`
    """

    def _is_satisfied(self) -> bool:
        timestamp_what = self.convert_timestamp_from_any(self.what)
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                # if self.value_is_none(i) and self.value_is_none(self.what):
                #     continue
                # elif self.value_is_none(i) or self.value_is_none(self.what):
                #     return True
                timestamp_i = self.convert_timestamp_from_any(i)
                # if not timestamp_i or not timestamp_what:
                #     return False
                if timestamp_what != timestamp_i:
                    return True
            return False
        else:
            for i in self.values:
                # if self.value_is_none(i) and self.value_is_none(self.what):
                #     return False
                # elif self.value_is_none(i) or self.value_is_none(self.what):
                #     continue
                timestamp_i = self.convert_timestamp_from_any(i)
                # if not timestamp_i or not timestamp_what:
                #     return False
                if timestamp_what == timestamp_i:
                    return False
            return True


class DateNeqSchema(DateConditionSchema):
    """
        JSON schema for not equals numeric condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return DateNeq(**data)



from .day_eq import MonthDayEqSchema
from .day_gt import MonthDayGtSchema
from .day_gte import MonthDayGteSchema
from .day_lt import MonthDayLtSchema
from .day_lte import MonthDayLteSchema
from .day_neq import MonthDayNeqSchema



"""
    String conditions base class
"""

import datetime
import logging

from dateutil.parser import parse
from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


class DateCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    class DateFormat:
        ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    def __init__(self, values, what, date_format=DateFormat.ISO_FORMAT, qualifier=ConditionBase.Qualifier.ForAnyValue,
                 **kwargs):
        self.values = values
        self.what = what
        self.qualifier = qualifier
        self.date_format = date_format

    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()

    def convert_string_to_timestamp(self, date_str):
        return round(
            datetime.datetime.strptime(date_str, self.date_format).replace(tzinfo=datetime.timezone.utc).timestamp())

    def convert_date_to_timestamp(self, d):
        return round(d.replace(tzinfo=datetime.timezone.utc).timestamp())

    def convert_timestamp_to_date(self, timestamp):
        # try:
        return datetime.datetime.fromtimestamp(timestamp)
        # except:
        #     return None

    def convert_string_to_timestamp_from_parse(self, date_str):
        try:
            return round(parse(date_str).replace(tzinfo=datetime.timezone.utc).timestamp())
        except:
            return None

    def convert_timestamp_from_any(self, i):
        if isinstance(i, (int, float)):
            timestamp_i = round(i)
        elif isinstance(i, str):
            timestamp_i = self.convert_string_to_timestamp_from_parse(i)
            if timestamp_i is None:
                timestamp_i = self.convert_string_to_timestamp(i)
        elif isinstance(i, datetime.datetime):
            timestamp_i = self.convert_date_to_timestamp(i)
        else:
            return False
        return self.convert_timestamp_to_round_minute(timestamp_i)

    def convert_timestamp_to_round_minute(self, value_timestamp):
        round_timestamp = round(value_timestamp)
        return round_timestamp - (round_timestamp % 60)


class DateConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.Raw(required=True, allow_none=True), required=True, allow_none=False)
    what = fields.Raw(required=True, allow_none=True)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                              validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    date_format = fields.String(allow_none=False, load_default=DateCondition.DateFormat.ISO_FORMAT, )

    class Meta:
        unknown = EXCLUDE



"""
    Numeric not equal conditions
"""

from marshmallow import post_load

from .base import NumericCondition, NumericConditionSchema


class Neq(NumericCondition):
    """
        Condition for number `what` not equals `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.what != i:
                    return True
            return False
        else:
            for i in self.values:
                if self.what == i:
                    return False
            return True


class NeqSchema(NumericConditionSchema):
    """
        JSON schema for not equals numeric condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Neq(**data)



"""
    String conditions base class
"""

import logging

from marshmallow import Schema, fields, EXCLUDE, validate

from ..base import ConditionBase, ABCMeta, abstractmethod

LOG = logging.getLogger(__name__)


class StringCondition(ConditionBase, metaclass=ABCMeta):
    """
        Base class for string conditions
    """

    def __init__(self, values, what, qualifier=ConditionBase.Qualifier.ForAnyValue, ignore_case=False,
                 delimiter="", **kwargs):
        # xử lý ignore case
        self.case_insensitive = ignore_case
        if ignore_case:
            self.values = [self.utf8_to_ascii(i) if isinstance(i, str) else i for i in values]
            self.what = self.utf8_to_ascii(what) if isinstance(what, str) else what
            self.delimiter = self.utf8_to_ascii(delimiter) if isinstance(delimiter, str) else delimiter
        else:
            self.values = [i for i in values]
            self.what = what
            self.delimiter = delimiter
        self.qualifier = qualifier

    @abstractmethod
    def _is_satisfied(self) -> bool:
        """
            Is string conditions satisfied

            :param what: string value to check
            :return: True if satisfied else False
        """
        raise NotImplementedError()

    @classmethod
    def is_string(cls, value) -> bool:
        """
            Check if value is string
        """
        return isinstance(value, str)


class StringConditionSchema(Schema):
    """
        Base JSON schema for string conditions
    """
    values = fields.List(fields.String(required=True, allow_none=True), required=True, allow_none=False)
    ignore_case = fields.Bool(default=False, missing=False)
    what = fields.String(required=True, allow_none=True)
    qualifier = fields.String(allow_none=False, load_default=ConditionBase.Qualifier.ForAnyValue,
                              validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    delimiter = fields.String(required=False, default="", missing="")

    class Meta:
        unknown = EXCLUDE



"""
    Numeric greater than conditions
"""

from marshmallow import post_load

from .base import NumericCondition, NumericConditionSchema


class Gt(NumericCondition):
    """
        Condition for number `what` greater than `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if self.what > i:
                    return True
            return False
        else:
            for i in self.values:
                if self.what <= i:
                    return False
            return True


class GtSchema(NumericConditionSchema):
    """
        JSON schema for greater than numeric condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Gt(**data)



"""
    String equals conditions
"""

from marshmallow import post_load, EXCLUDE

from .base import StringCondition, StringConditionSchema


class Equals(StringCondition):
    """
        Condition for string `self.what` equals `value`
    """

    def _is_satisfied(self) -> bool:
        if self.qualifier == self.Qualifier.ForAnyValue:
            for i in self.values:
                if not self.is_string(i) and not self.is_string(self.what):
                    return True
                elif not self.is_string(i) or not self.is_string(self.what):
                    continue

                if self.what == i:
                    return True
            return False
        else:
            for i in self.values:
                if not self.is_string(i) and not self.is_string(self.what):
                    continue
                elif not self.is_string(i) or not self.is_string(self.what):
                    return False

                if self.what != i:
                    return False
            return True


class EqualsSchema(StringConditionSchema):
    """
        JSON schema for equals string condition
    """

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        # self.validate(data)
        return Equals(**data)




"""
    Operation base class
"""

from abc import ABCMeta, abstractmethod
import re

# from py_abac.context import EvaluationContext
# what is value of field, values is list value policy
MapAnyValue = ""

patterns = {
    '[àáảãạăắằẵặẳâầấậẫẩ]': 'a',
    '[đ]': 'd',
    '[èéẻẽẹêềếểễệ]': 'e',
    '[ìíỉĩị]': 'i',
    '[òóỏõọôồốổỗộơờớởỡợ]': 'o',
    '[ùúủũụưừứửữự]': 'u',
    '[ỳýỷỹỵ]': 'y'
}


class ConditionBase(metaclass=ABCMeta):
    """
        Base class for conditions
    """

    # @abstractmethod
    # def is_satisfied(self, ctx: EvaluationContext) -> bool:
    #     """
    #         Is conditions satisfied?
    #
    #         :param ctx: evaluation context
    #         :return: True if satisfied else False
    #     """
    #     raise NotImplementedError()

    class Qualifier:
        ForAnyValue = "ForAnyValue"
        ForAllValues = "ForAllValues"
        ALL = [ForAnyValue, ForAllValues]

    @classmethod
    def value_is_none(cls, value):
        if not value and value not in [0, False]:
            return True
        return False

    @classmethod
    def utf8_to_ascii(cls, text):
        if not isinstance(text, str):
            return ''
        output = text
        for regex, replace in patterns.items():
            output = re.sub(regex, replace, output)
            output = re.sub(regex.upper(), replace.upper(), output)
        return output.lower()



"""
    Policy class
"""
import json
from marshmallow import Schema, fields, post_load, ValidationError, validate, EXCLUDE
from .conditions.schema import ConditionSchema
from .utils import Utils
from .conditions.base import ConditionBase
from .exceptions import *


class AccessType:
    DENY_ACCESS = "deny"
    ALLOW_ACCESS = "allow"


class Policy(object):
    """
        Policy class containing rules and targets
    """

    def __init__(
            self,
            effect: str,
            # action: list,
            # resource: list,
            condition: dict,
            request_access: dict, **kwargs
    ):
        self.effect = effect
        # self.action = action
        # self.resource = resource
        self.condition = condition
        self.condition_convert = None
        self.request_access = request_access

    def is_allowed(self):
        conditions = self.convert_condition()
        self.condition_convert = conditions
        for condition in conditions:
            # print("abac_sdk condition: ", condition)
            if condition.get("operator") == "Null":
                what_check = self.get_value_from_field(condition.get("field"))
                if what_check is not None:
                    return False
            else:
                cond_schema = ConditionSchema().load(condition)
                if not cond_schema._is_satisfied():
                    return False
        return True

    def get_condition_convert(self):
        return self.condition_convert

    def convert_condition(self):
        """
            {
                "operator": "StringEquals",
                "field": "user:staff_code",
                "values": ["A123456"],
                "qualifier": "ForAnyValue",
                "if_exists": false,
                "ignore_case": false,
            }
        :return:
        """
        list_condition = []
        have_condition_exclude = False
        for item in self.condition:
            cond_schema = self.validate_item_condition(item)
            # print("abac_sdk cond_schema: {}".format(json.dumps(cond_schema)))
            if cond_schema.get("operator") == "Null":
                list_condition.append({**item})
            elif cond_schema.get("operator") in ["Exists", "NotExists"]:
                if_exists = cond_schema.get("if_exists")
                what_check = self.get_value_from_field(cond_schema.get("field"))
                if if_exists:
                    if not what_check and what_check not in [0, False]:
                        continue
                resource_name, resource_key = Utils.split_delemiter_resource(cond_schema.get("field"))
                data_resource = {}
                if resource_name:
                    if self.request_access.get(resource_name) and isinstance(
                            self.request_access.get(resource_name), dict):
                        data_resource = self.request_access.get(resource_name)
                list_condition.append({
                    **item,
                    "values": data_resource,
                    "what": resource_key,
                })
            else:
                if_exists = cond_schema.get("if_exists")
                what_check = self.get_value_from_field(cond_schema.get("field"))
                if if_exists:
                    if not what_check and what_check not in [0, False]:
                        continue
                # if what_check is None:
                #     raise GetValueNoneException("{} get value is None".format(cond_schema.get("field")))
                if not what_check and isinstance(what_check, str):
                    what_check = None
                values = []
                for v in cond_schema.get("values"):
                    value_from_variable = self.get_value_from_variable(v)
                    if not value_from_variable and isinstance(value_from_variable, str):
                        value_from_variable = None
                    if isinstance(value_from_variable, list):
                        values.extend(value_from_variable)
                    else:
                        values.append(value_from_variable)
                list_condition.append({
                    **item,
                    "values": values,
                    "what": what_check,
                })
        # if len(list_condition) == 0 and not have_condition_exclude:
        #     raise EmptyConditionException("no condition found")
        return list_condition

    def get_value_from_variable(self, str_variable):
        if isinstance(str_variable, str) and Utils.check_field_is_variable(str_variable):
            variables = Utils.get_field_key_from_variable(str_variable)
            if variables:
                if len(variables) == 1:
                    field_value = self.get_value_from_field(variables[0])
                    if field_value is None:
                        # raise GetValueNoneException("{} get value is None".format(variables[0]))
                        return field_value
                    if isinstance(field_value, str):
                        field_value_format = Utils.replace_variable_to_value(variables[0], field_value, str_variable)
                        return field_value_format
                    return field_value
                for field_key in variables:
                    field_value = self.get_value_from_field(field_key)
                    if field_value is None:
                        # raise GetValueNoneException("{} get value is None".format(field_key))
                        field_value = ""
                    field_value = str(field_value)
                    str_variable = Utils.replace_variable_to_value(field_key, field_value, str_variable)
        return str_variable

    def get_value_from_field(self, field_key):
        resource_name, resource_key = Utils.split_delemiter_resource(field_key)
        if resource_name and resource_key:
            if self.request_access.get(resource_name) and isinstance(self.request_access.get(resource_name), dict):
                data_resource = self.request_access.get(resource_name)
                value = Utils.get_nested_value(data_resource, resource_key)
                return value
        return None

    def validate_item_condition(self, obj):
        try:
            return ConditionValidateSchema().load(obj)
        except Exception as err:
            raise InvalidConditionException("validate_item_condition: {}".format(err))


class ConditionValidateSchema(Schema):
    operator = fields.String(required=True, allow_none=False)
    field = fields.String(required=True, allow_none=False)
    values = fields.List(fields.Raw(required=True, allow_none=False), required=True, allow_none=False)
    qualifier = fields.String(default=ConditionBase.Qualifier.ForAnyValue,
                              validate=validate.OneOf(ConditionBase.Qualifier.ALL))
    if_exists = fields.Boolean(default=False)
    ignore_case = fields.Boolean(default=False)

    class Meta:
        unknown = EXCLUDE


class PolicySchema(Schema):
    """
        JSON schema for policy
    """
    effect = fields.String(required=True, validate=validate.OneOf([AccessType.DENY_ACCESS, AccessType.ALLOW_ACCESS]))
    # action = fields.List(fields.String(required=True, allow_none=False),required=True, allow_none=False)
    # resource = fields.List(fields.String(required=True, allow_none=False),required=True, allow_none=False)
    condition = fields.List(fields.Raw(required=True, allow_none=False), required=True, allow_none=False)
    request_access = fields.Dict(default={}, missing={}, allow_none=False)

    class Meta:
        unknown = EXCLUDE

    @post_load
    def post_load(self, data, **_):  # pylint: disable=missing-docstring,no-self-use
        return Policy(**data)
